/* Copyright (C) 2003 Nigel A. Hathaway.  All rights reserved.
  
 The contents of this file are subject to the wxWindows License
 Version 3.0 (the "License"); you may not use this file except in
 compliance with the License. You may obtain a copy of the License at
 http://www.wxwindows.org/licence3.txt

 Software distributed under the License is distributed on an
 "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 implied. See the License for the specific language governing
 rights and limitations under the License.

*/

/* $Id: ghostscript.cpp,v 1.3 2003/09/01 15:49:54 nhathaway Exp $ */

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#ifdef __GNUG__
    #pragma implementation "ghostscript.h"
#endif

#include "wx/wxprec.h"

#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#include "wx/config.h"
#include "ghostscript.h"

#ifdef __WXGTK__
#include <gdk/gdk.h>
#endif

#ifdef __UNIX__
#include <sched.h>
#endif

IMPLEMENT_CLASS(wxGsDll, wxObject)

// Ghostscript DLL constructor
wxGsDll::wxGsDll(wxString GsDllLocation, wxString GsLibLocation, wxString GsFontsLocation)
{
    // Initialise instance data with defaults.
    m_OK = false;
    m_version = 0.0;
    if (GsDllLocation.IsEmpty())
        GsDllLocation = ms_defaultDllLocation;
    m_GsLibLocation = GsLibLocation;
    m_GsFontsLocation = GsFontsLocation;

    // Link to the DLL
    if (!m_dll.Load(GsDllLocation))
        return;

    // Load the pointers to routines
    {
        wxLogNull logNo; // prevent message(s)
        gsapi_revision = (gs::PFN_gsapi_revision)m_dll.GetSymbol(wxT("gsapi_revision"));
        gsapi_new_instance = (gs::PFN_gsapi_new_instance)m_dll.GetSymbol(wxT("gsapi_new_instance"));
        gsapi_delete_instance = (gs::PFN_gsapi_delete_instance)m_dll.GetSymbol(wxT("gsapi_delete_instance"));
        gsapi_set_stdio = (gs::PFN_gsapi_set_stdio)m_dll.GetSymbol(wxT("gsapi_set_stdio"));
        gsapi_set_poll = (gs::PFN_gsapi_set_poll)m_dll.GetSymbol(wxT("gsapi_set_poll"));
        gsapi_set_display_callback = (gs::PFN_gsapi_set_display_callback)m_dll.GetSymbol(wxT("gsapi_set_display_callback"));
        gsapi_init_with_args = (gs::PFN_gsapi_init_with_args)m_dll.GetSymbol(wxT("gsapi_init_with_args"));
        gsapi_run_string_begin = (gs::PFN_gsapi_run_string_begin)m_dll.GetSymbol(wxT("gsapi_run_string_begin"));
        gsapi_run_string_continue = (gs::PFN_gsapi_run_string_continue)m_dll.GetSymbol(wxT("gsapi_run_string_continue"));
        gsapi_run_string_end = (gs::PFN_gsapi_run_string_end)m_dll.GetSymbol(wxT("gsapi_run_string_end"));
        gsapi_run_string_with_length = (gs::PFN_gsapi_run_string_with_length)m_dll.GetSymbol(wxT("gsapi_run_string_with_length"));
        gsapi_run_string = (gs::PFN_gsapi_run_string)m_dll.GetSymbol(wxT("gsapi_run_string"));
        gsapi_run_file = (gs::PFN_gsapi_run_file)m_dll.GetSymbol(wxT("gsapi_run_file"));
        gsapi_exit = (gs::PFN_gsapi_exit)m_dll.GetSymbol(wxT("gsapi_exit"));
        gsapi_set_visual_tracer = (gs::PFN_gsapi_set_visual_tracer)m_dll.GetSymbol(wxT("gsapi_set_visual_tracer"));
    }

    // Check they all resolved OK (visual tracer is allowed not to be present)
    if (!(gsapi_revision && gsapi_new_instance && gsapi_delete_instance && gsapi_set_stdio && gsapi_set_poll &&
          gsapi_set_display_callback && gsapi_init_with_args && gsapi_run_string_begin && gsapi_run_string_continue &&
          gsapi_run_string_end && gsapi_run_string_with_length && gsapi_run_string && gsapi_run_file && gsapi_exit))
         return;

    // Get the version (first DLL call into Ghostscript - tests the DLL link worked OK)
    gs::gsapi_revision_t rev;

    if (gsapi_revision(&rev, sizeof(rev)) == 0)
        m_version = ((double)rev.revision)/100;
    else
        return;
	
    m_OK = true;
}

wxGsDll::~wxGsDll()
{
    // Unload the DLL, if there was one
    if (m_dll.IsLoaded())
        m_dll.Unload();
}

// Activates the visual tracer (see GS docs)
void wxGsDll::SetVisualTracer(wxGsVisualTracer& vd)
{
    // If present in the API, set it up
    if (gsapi_set_visual_tracer)
        gsapi_set_visual_tracer(&vd);
}

// Platform-specific default path to GS DLL.
void wxGsDll::InitDefaults()
{
#ifdef __WIN32__
    bool notfound = false;

    // Look for best Ghostscript registry entry
    wxRegKey regkey(wxT("HKEY_LOCAL_MACHINE\\SOFTWARE"));
    if (regkey.HasSubKey(wxT("GPL Ghostscript")))
        regkey.SetName(regkey.GetName(FALSE) + wxT("\\GPL Ghostscript"));
    else if (regkey.HasSubKey(wxT("AFPL Ghostscript")))
        regkey.SetName(regkey.GetName(FALSE) + wxT("\\AFPL Ghostscript"));
    else if (regkey.HasSubKey(wxT("GNU Ghostscript")))
        regkey.SetName(regkey.GetName(FALSE) + wxT("\\GNU Ghostscript"));
    else
        notfound = true;

    // Find most recent version
    if (!notfound)
    {
        long l;
        wxString s, k;
        double v=0.0;
        if (regkey.GetFirstKey(s, l))
        {
            do
            {
                double d;
                if (s.ToDouble(&d))
                {
                    if (d > v)
                    {
                        k = s;
                        v = d;
                    }
                }
            } while(regkey.GetNextKey(s, l));
        }

        if (v==0.0)
            notfound = true;
        else
            regkey.SetName(regkey.GetName(FALSE) + wxT("\\") + k);
    }

    // Find the DLL location
    if (!notfound)
    {
        wxString loc;
        if (regkey.QueryValue(wxT("GS_DLL"), loc))
            ms_defaultDllLocation = loc;
        else
            notfound = true;
    }

    if (notfound)
        ms_defaultDllLocation = wxString(wxT("gsdll32.dll"));

#elif defined(__WXOS2__)
#error OS/2 port not yet written. Feel free to contribute.
#elif defined(__UNIX__)
   if (wxFileName::IsFileReadable(wxT("/usr/local/lib/libgs.so")))
      ms_defaultDllLocation = wxString(wxT("/usr/local/lib/libgs.so"));
   else if (wxFileName::IsFileReadable(wxT("/usr/local/lib/libgs.so.8")))
      ms_defaultDllLocation = wxString(wxT("/usr/local/lib/libgs.so.8"));
   else if (wxFileName::IsFileReadable(wxT("/usr/lib/libgs.so"))) // this was default
      ms_defaultDllLocation = wxString(wxT("/usr/lib/libgs.so"));
   else if (wxFileName::IsFileReadable(wxT("/usr/lib/libgs.so.8")))
      ms_defaultDllLocation = wxString(wxT("/usr/lib/libgs.so.8"));
#elif defined(__WXMAC__)
# ifdef __DARWIN__
    ms_defaultDllLocation = wxString(wxT("/sw/lib/libgs.dylib")); // Fink, else TBD
# elif TARGET_CARBON
    ms_defaultDllLocation = wxString(wxT("GhostscriptLib Carbon"));
# else
    ms_defaultDllLocation = wxString(wxT("GhostscriptLib PPC"));
# endif
#else
#error Platform not supported.
#endif
}

wxString wxGsDll::ms_defaultDllLocation;

IMPLEMENT_DYNAMIC_CLASS(wxGhostscript, wxObject)

// Defaults for uninitialized Ghostscript instance
void wxGhostscript::Init()
{
    m_gsdll = NULL;
    m_imagetype = none;
    m_instance = NULL;
    m_exitcode = 0;
    m_thread = NULL;
    m_callbackswitch = false;
    m_phase = m_resume = uninitialized;
}

// Set up the Ghostscript instance so it is ready to run
bool wxGhostscript::Create(wxGsDll *gsdll, const wxArrayString& argarray, wxGhostscript::ImageType imagetype)
{
    Init();  // Null/zero defaults

    wxCHECK_MSG(gsdll && gsdll->isOK(), false, wxT("bad ghostscript DLL"));

    // Create the Ghostscript thread and set it going
    m_gsdll = gsdll;
    m_thread = new wxGsThread(m_stdinbuf, this, wxTHREAD_JOINABLE);

    wxCHECK_MSG(m_thread, false, wxT("unable to create ghostscript thread"));
    wxCHECK_MSG(m_thread->m_mutex[0].Lock() == wxMUTEX_NO_ERROR &&
                m_thread->Create() == wxTHREAD_NO_ERROR &&
                m_thread->Run() == wxTHREAD_NO_ERROR,
                false, wxT("unable to initialize ghostscript thread"));

    // Get the thread into the waiting state
    m_phase = before;
    while (m_thread->m_mutex[1].TryLock() != wxMUTEX_BUSY)
    {
        m_thread->m_mutex[1].Unlock();
        wxThread::This()->Yield();
    }
    m_thread->m_mutex[2].Lock();
    m_thread->m_unowned = 1;

    // Create the Ghostscript instance, and link in the callbacks
    wxCHECK_MSG(m_gsdll->gsapi_new_instance(&m_instance, this) == 0, false, wxT("can't create ghostscript instance"));

    if (m_gsdll->gsapi_set_stdio(m_instance, static_stdin, static_stdout, static_stderr) != 0 ||
        m_gsdll->gsapi_set_poll(m_instance, static_poll) != 0 ||
        m_gsdll->gsapi_set_display_callback(m_instance, &m_display_callback) != 0)
    {
        Terminate(e_Fatal);
        m_gsdll->gsapi_delete_instance(m_instance);
        wxFAIL_MSG(wxT("can't set ghostscript callbacks"));
        return false;
    }


    // Set up the image type
    m_imagetype = imagetype;
    if (m_imagetype == separations && m_gsdll->m_version < 9999) // for now, until separations are supported
        m_imagetype = greyscale;

    // Process the provided arguments, ensuring they are made safe
    wxArrayString args;
    wxString libarg, fontarg, devicearg;

    // Filter out any attempts to specify our special arguments
    for (size_t  i=0; i<argarray.GetCount(); i++)
    {
        wxString arg = argarray[i];
        if (!(arg.StartsWith(wxT("-I"), &libarg) ||
              arg.StartsWith(wxT("-sFONTPATH="), &fontarg) ||
              arg.StartsWith(wxT("-sDEVICE="), &devicearg) ||
              arg.StartsWith(wxT("-dDisplayHandle=")) ||
              arg.StartsWith(wxT("-dDisplayFormat="))))
            args.Add(arg);
    }

    // Put back in our versions of those arguments
    if (libarg.IsEmpty() && (!m_gsdll->m_GsLibLocation.IsEmpty() || !m_gsdll->m_GsFontsLocation.IsEmpty()))
    {
        libarg = m_gsdll->m_GsLibLocation;
        if (!libarg.IsEmpty() && !m_gsdll->m_GsFontsLocation.IsEmpty())
        {
#if defined(__WXMSW__) || defined(__WXOS2__)
            libarg << wxT(";");
#elif defined(__WXMAC__)
            libarg << wxT(",");
#else // Unix
            libarg << wxT(":");
#endif
        }
        libarg << m_gsdll->m_GsFontsLocation;
    }

    if (!libarg.IsEmpty())
        args.Insert(wxT("-I") + libarg, 0);

    if (fontarg.IsEmpty() && !m_gsdll->m_GsFontsLocation.IsEmpty())
        fontarg = m_gsdll->m_GsFontsLocation;

    if (!fontarg.IsEmpty())
        args.Insert(wxT("-sFONTPATH=") + fontarg, 0);

    // The output device defaults to the display
    if (devicearg.IsEmpty())
        devicearg = wxT("display");
    if (m_imagetype == none && devicearg == wxT("display"))
        devicearg = wxT("nullpage");
    if (devicearg != wxT("display"))
        m_imagetype = none;

    // Set up the device parameters
    args.Insert(wxT("-sDEVICE=") + devicearg, 0);
    if (devicearg == wxT("display"))
     {
    args.Insert(wxString::Format(wxT("-sDisplayHandle=%ld"), (long)this), 0);
    args.Insert(wxString::Format(wxT("-dDisplayFormat=%d"),
        (m_imagetype == colour ? wxGsImage::ms_colourfmt : wxGsImage::ms_greyscalefmt)),0);
     }
    args.Insert(wxT("gs"), 0);

    // Convert C++ style arg array to C-style arga array
    int argc = args.GetCount();
    wxWritableCharBuffer *argb = new  wxWritableCharBuffer[argc];
    char **argv = new char*[argc];
    for (int j=0; j<argc; j++)
     {
//        argv[j] = (char *)(args[j].c_str());
        argb[j] = args[j].char_str(wxConvISO8859_1); // PSt version
	argv[j] = argb[j];
     }

    // Initialiase the Ghostscript instance witht the command-line parameters
    m_exitcode = m_gsdll->gsapi_init_with_args(m_instance, argc, argv);
    if (!(m_exitcode == 0 || m_exitcode == e_NeedInput))
        Terminate(m_exitcode);

    delete []argv;
    delete []argb;

    m_callbackswitch = true;
    return m_phase != uninitialized;
}

wxGhostscript::~wxGhostscript()
{
    // If the Ghostscript thread hasn't already exited, make it do so
    if (m_phase != uninitialized)
        Terminate(e_Quit);

    m_callbackswitch = false;

    // Close down the thread and its facilities
    if (m_thread)
    {
        m_thread->m_mutex[0].Unlock();
        m_thread->m_mutex[1].Unlock();
        m_thread->m_mutex[2].Unlock();
        wxMilliSleep(10); // wait without processing event loop
//        m_thread->Wait();
        delete m_thread;
    }
}

// Do the required operation, passing the data to the other thread
bool wxGhostscript::DoOperation(wxString data, GsOperation op)
{
    wxCHECK_MSG(m_thread, false, wxT("no ghostscript thread"));
    wxCHECK_MSG(isReady(), false, wxT("ghostscript in wrong state"));

    // Set up the parameters of the operation
    m_thread->m_data = data;
    m_thread->m_op = op;

    // Do the operation, processing any callback operations that may occur
    while (TRUE)
    {
        m_thread->m_callbackop = nocall;

        RunOtherThread();

        if (!DoCallbackOp())
            break;
    }

    return true;
}

// Callback operation as called from GS itself
int wxGhostscript::DoCallback(wxGsImage *image, wxString text, int copies, bool flush, wxRect region, GsCallbackOperation op)
{
    // Set up the parameters for passing to the other thread
    m_thread->m_image = image;
    m_thread->m_text = text;
    m_thread->m_copies = copies;
    m_thread->m_flush = flush;
    m_thread->m_region = region;
    m_thread->m_callbackop = op;

    // GS is initialised in the context of the main thread, so only run the other thread
    // once initialisation is complete, otherwise just proces the callback straight from here.
    if (m_callbackswitch)
        RunOtherThread();
    else
        DoCallbackOp();

    return m_thread->m_retval;
}

// Perform the callbacks by calling the virtual functions.
// These calls always occur within the context of the main thread
//   so that GDI calls can be safely made from them.
bool wxGhostscript::DoCallbackOp()
{
    bool retval = true;
    wxString *textin;

    switch (m_thread->m_callbackop)
    {
    case onstdin:
        m_thread->m_retval = 0;
        textin = OnStdin();
        if (textin)
            m_thread->m_textin = textin;  // Never a NULL momemt ;-)
        break;

    case onstdout:
        m_thread->m_retval = 0;
        OnStdout(m_thread->m_text);
        break;

    case onstderr:
        m_thread->m_retval = 0;
        OnStderr(m_thread->m_text);
        break;

    case onpoll:
        m_thread->m_retval = OnPoll();
        break;

    case onopen:
        m_thread->m_retval = OnDisplayOpen(*(m_thread->m_image));
        break;

    case onpreclose:
        m_thread->m_retval = OnDisplayPreClose(*(m_thread->m_image));
        break;

    case onclose:
        m_thread->m_retval = OnDisplayClose(*(m_thread->m_image));
        break;

    case onpresize:
        m_thread->m_retval = OnDisplayPreSize(*(m_thread->m_image));
        break;

    case onsize:
        m_thread->m_retval = OnDisplaySize(*(m_thread->m_image));
        break;

    case onsync:
        m_thread->m_retval = OnDisplaySync(*(m_thread->m_image));
        break;

    case onpage:
        m_thread->m_retval = OnDisplayPage(*(m_thread->m_image), m_thread->m_copies, m_thread->m_flush);
        break;

    case onupdate:
        m_thread->m_retval = OnDisplayUpdate(*(m_thread->m_image), m_thread->m_region);
        break;

    default:
        retval = false;
    }

    return retval;
}

// The main GS thread function
void wxGhostscript::GsThread()
{
    // 2 out of the 3 mutexes are owned when running
    m_thread->m_mutex[2].Lock();
    m_thread->m_mutex[1].Lock();

    // Loop until told to terminate
    while (true)
    {
        RunOtherThread();

        // Local copy of the function parameter
        wxString str = m_thread->m_data;
        int result;

        switch (m_thread->m_op)
        {
        // Run a string in its own logical file context
        case run:
            m_phase = during;

            // Large strings have to be fed piecemeal into Ghostscript
            if (str.Length() > 65535)
            {
                // Set up the feed operation
                result = m_gsdll->gsapi_run_string_begin(m_instance, 0, &m_exitcode);

                if (!(m_exitcode == 0 || m_exitcode == e_NeedInput))
                {
                    m_phase = after;
                    goto quit;
                }

                // Set up the feed loop parameters
                size_t length = 65535;
                const char *c_str = str.mb_str(wxConvISO8859_1);
                size_t remaining = str.Length();

                while (remaining)
                {
                    // Actually feed the string to Ghostscript.
                    result = m_gsdll->gsapi_run_string_continue(m_instance, c_str, length, 0, &m_exitcode);

                    // This happens if Ghostscript wants to terminate
                    if (result != e_NeedInput)
                    {
                        m_phase = after;

                        if (m_exitcode == e_Quit)
                            result = m_gsdll->gsapi_run_string_end(m_instance, 0, &m_exitcode);

                        goto quit;
                    }

                    // Next iteration of the feed
                    remaining -= length;
                    c_str += length;
                    if (length > remaining)
                        length = remaining;
                }

                result = m_gsdll->gsapi_run_string_end(m_instance, 0, &m_exitcode);
            }
            else
                result = m_gsdll->gsapi_run_string_with_length(m_instance, (const char*)str.mb_str(wxConvISO8859_1), str.Length(), 0, &m_exitcode);

            if (result < 0)
            {
                m_phase = after;
                goto quit;
            }
            m_phase = before;
            break;

        // Run a file on its own
        case file:
            m_phase = during;
            result = m_gsdll->gsapi_run_file(m_instance, (const char*)str.fn_str(), 0, &m_exitcode);
            if (result < 0)
            {
                m_phase = after;
                goto quit;
            }
            m_phase = before;
            break;

        // Exit Ghostscript and the Ghostscript thread
        case terminate:
            if (m_phase != before)
                m_phase = after;
            goto quit;
        }
    }

quit:

    // Clean-up operation
    m_gsdll->gsapi_exit(m_instance);
    m_gsdll->gsapi_delete_instance(m_instance);
    m_phase = uninitialized;

    // Release owned mutexes
    for (int i=0; i<3; i++)
    {
        if (i == m_thread->m_unowned)
            continue;

        m_thread->m_mutex[i].Unlock();
    }
}

// Request suspension of Ghostscript
// (called only from callbacks, causing the main operation to return)
// Suspension only happens once the virtual function has returned.
bool wxGhostscript::Suspend()
{
    if (m_phase == suspended)
        return true;

    if (m_phase == uninitialized)
        return false;

    m_resume = m_phase;
    m_phase = suspended;

    return true;
}

// Trigger termination of Ghostscript
bool wxGhostscript::Terminate(int errorcode)
{
    m_thread->m_retval = m_thread->m_errorcode = errorcode;

    while (m_phase == suspended)
        Resume();

    if (m_phase == uninitialized)
        return true;

    return DoOperation(wxEmptyString, terminate);
}

// Resume GS processing after it has been suspended.
// This will return either when the suspended GS operation is
//   complete or when suspend is invoked again
bool wxGhostscript::Resume()
{
    if (m_phase != suspended)
        return false;

    m_phase = m_resume;

    // Continue processing the original operation and callbacks
    while (TRUE)
    {
        m_thread->m_callbackop = nocall;

        RunOtherThread();

        if (!DoCallbackOp())
            break;
    }

    return true;
}

// Called from within the real GS callbacks to cause suspension
//  to actually happen when it has been requested
int wxGhostscript::CheckSuspended(int code)
{
    // Suspend if request and there are no errors
    if (code >= 0 && m_phase == suspended)
    {
        RunOtherThread();

        if (m_thread->m_errorcode < 0)
            code = m_thread->m_errorcode;
    }

    return code;
}

// Do a co-operative thread switch
void wxGhostscript::RunOtherThread()
{
    // This involves juggling with 3 balls, and is done this way
    //   so as to be guaranteed to work on all platforms.
    int unowned = m_thread->m_unowned;
    int owned = unowned - 1;
    if (owned < 0)
        owned = 2;

    // Pass one ball along and wait to catch the other one
    m_thread->m_mutex[owned].Unlock();
    m_thread->m_mutex[unowned].Lock();

    m_thread->m_unowned = owned;
}

// Real Ghostscript callbacks
gs::display_callback wxGhostscript::m_display_callback =
{
    sizeof (gs::display_callback),  /* Size of this structure */
    DISPLAY_VERSION_MAJOR,          /* Major version of this structure  */
    DISPLAY_VERSION_MINOR,          /* Minor version of this structure */
    &static_display_open,            /* New device has been opened */
    &static_display_preclose,        /* Device is about to be closed. */
    &static_display_close,           /* Device has been closed. */
    &static_display_presize,         /* Device is about to be resized. */
    &static_display_size,            /* Device has been resized. */
    &static_display_sync,            /* flushpage */
    &static_display_page,            /* showpage */
    &static_display_update,          /* Notify the caller whenever a portion of the raster is updated. */
    &static_display_memalloc,        /* Allocate memory for bitmap */
    &static_display_memfree           /* Free memory for bitmap */
#if DISPLAY_VERSION_MAJOR > 1
    , NULL // static_display_separation        /* Special color was set */
#endif
};

// Process GS's stdin request
int wxGhostscript::static_stdin(void *caller_handle, char *buf, int len)
{
    wxGhostscript *gs = wxDynamicCast(caller_handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    // Feed back any leftovers from last time before asking for more
    if (gs->m_stdinbuf.IsEmpty())
    {
        // The callback gets tells the system that data is needed.
        // The buffer is filled while GS is suspended
        int susp = gs->CheckSuspended(gs->DoCallback(NULL, onstdin));
        if (susp < 0)
            return susp;

        // Copy the input buffer data over
        gs->m_stdinbuf = *(gs->m_thread->m_textin);
    }

    // Adjust fir the case where the data provided is less than that requested
    if ((size_t)len > gs->m_stdinbuf.Length())
        len = gs->m_stdinbuf.Length();

    // Copy to GS the amount of data requested, and keep the leftovers
    memcpy(buf, (const char*)gs->m_stdinbuf.mb_str(wxConvISO8859_1), len);
    gs->m_stdinbuf.Remove(0, len);

    return len;
}

// Process GS's stdout request
int wxGhostscript::static_stdout(void *caller_handle, const char *str, int len)
{
    wxGhostscript *gs = wxDynamicCast(caller_handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    // Invoke the corresponding virtual function
    int susp = gs->CheckSuspended(gs->DoCallback(NULL, wxString(str, wxConvISO8859_1, len), onstdout));

    return susp < 0 ? susp : len;
}

// Process GS's stderr request
int wxGhostscript::static_stderr(void *caller_handle, const char *str, int len)
{
    wxGhostscript *gs = wxDynamicCast(caller_handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    // Invoke the corresponding virtual function
    int susp = gs->CheckSuspended(gs->DoCallback(NULL, wxString(str, wxConvISO8859_1, len), onstderr));

    return susp < 0 ? susp : len;
}

// Process GS's poll request
int wxGhostscript::static_pollcnt = CALLBACK_POLL_DIVIDER;

int wxGhostscript::static_poll(void *caller_handle)
{
   if (--static_pollcnt == 0) {
      static_pollcnt=CALLBACK_POLL_DIVIDER;
      wxGhostscript *gs = wxDynamicCast(caller_handle, wxGhostscript);
      wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));
      // Invoke the corresponding virtual function
      return gs->CheckSuspended(gs->DoCallback(NULL, onpoll));
   } else {
      return 0;
   }
}

// Process GS's display open request
int wxGhostscript::static_display_open(void *handle, void *device)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    if (!image)
    {
        image = new wxGsImage(device);
        gs->m_imagehash[device] = image;
    }

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, onopen));
}

// Process GS's display preclose request
int wxGhostscript::static_display_preclose(void *handle, void *device)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, onpreclose));
}

// Process GS's display close request
int wxGhostscript::static_display_close(void *handle, void *device)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Invoke the corresponding virtual function
    int code = gs->CheckSuspended(gs->DoCallback(image, onclose));
    gs->m_imagehash.erase(device);
    delete image;
    return code;
}

// Process GS's display pre-size request
int wxGhostscript::static_display_presize(void *handle, void *device, int width, int height, int raster, unsigned int format)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    wxCHECK_MSG(format==wxGsImage::ms_colourfmt || format==wxGsImage::ms_greyscalefmt,
                e_Fatal, wxT("invalid requested format"));

    image->InitParams(width, height, raster, format==wxGsImage::ms_colourfmt ? colour : greyscale);

    // Ultimately, with separations, this needs to be the name of the colour
    image->SetName(format==wxGsImage::ms_colourfmt ? wxT("color") : wxT("grayscale"));

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, onpresize));
}

// Process GS's display size request
int wxGhostscript::static_display_size(void *handle, void *device, int width, int height, int raster, unsigned int format, unsigned char *pimage)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, onsize));
}

// Process GS's display sync request
int wxGhostscript::static_display_sync(void *handle, void *device)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, onsync));
}

// Process GS's display page request
int wxGhostscript::static_display_page(void *handle, void *device, int copies, int flush)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Invoke the corresponding virtual function
    return gs->CheckSuspended(gs->DoCallback(image, copies, flush!=0, onpage));
}

// Process GS's display update request
int wxGhostscript::static_display_update(void *handle, void *device, int x, int y, int w, int h)
{
   static int xb=65535,yb=65535,xe=0,ye=0;
   if (xb>x)   xb=x;    if (yb>y)   yb=y;
   if (xe<x+w) xe=x+w;  if (xe<y+h) ye=y+h;
   if (--static_pollcnt == 0) {
      static_pollcnt=CALLBACK_POLL_DIVIDER;
      wxRect rect = wxRect(xb, yb, xe-xb, ye-yb);
      xb=65535,yb=65535,xe=0,ye=0;
      wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);
      wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));
      wxGsImage *image = gs->m_imagehash[device];
      wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));
      // Invoke the corresponding virtual function
      return gs->CheckSuspended(gs->DoCallback(image, rect, onupdate)); // top/bottom coordinate sys match ?
   } else {
      return 0;
   }
}

// Process GS's display memory allocate request
void *wxGhostscript::static_display_memalloc(void *handle, void *device, unsigned long size)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, NULL, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, NULL, wxT("invalid device from ghostscript"));

    // Allocate memory for the image
    return image->AllocateImage(size);
}

// Process GS's display memory free request
int wxGhostscript::static_display_memfree(void *handle, void *device, void *mem)
{
    wxGhostscript *gs = wxDynamicCast(handle, wxGhostscript);

    wxCHECK_MSG(gs, e_Fatal, wxT("invalid handle from ghostscript"));

    wxGsImage *image = gs->m_imagehash[device];

    wxCHECK_MSG(image, e_Fatal, wxT("invalid device from ghostscript"));

    // Do not free the image in case it is required elsewhere (freed on close or re-allocate)
    return 0;
}

// Porting notes for the platform-specific image class
// ---------------------------------------------------

/* Firstly, some important points to note:
 *
 * 1. GS generates bitmaps that can be directly transferred to graphics memory
 *    without processing (well, nearly - it'll be there one day)
 * 2. wxWindows does not have the facility to deal in memory addressable images
 *    and transfer then to the display in an efficient manner.
 * 3. GS can generate very large bitmaps and certain platforms - notably Windows
 *    and Mac - are very quirky when it come to handling large bitmaps.
 * 4. We need to optimize for speed of data transfer in case the application
 *    wants to do progressive display updates as the image is rendering.
 *
 * Ideally we want the OS to allocate memory on the graphics card wherever possible.
 * Failing that, we at least need to be in the same format as the graphics card (and
 * its current mode). GS doesn't support all modes yet: e.g. 128-bit card where the
 * rowstride needs to be a multiple of 128 bits.
 *
 * So on Windows we ideally want to use DirectX. On Mac, the OS will allocate a native
 * format GWorld for us, and process it efficiently. On X we want to use as a minimum
 * the shared memory extension for cases where the X-server is on the host, and possibly
 * even use direct drawing extensions where available.
 *
 * As can be seen, these ideals have not yet been implemented. The complexities have not
 * been tackled and the easy route has been taken in the short term. In general, what we
 * have at the moment is a fallback in case facilities are not available. A more developed
 * version would pick the most efficient method depending on what hardware/OS/GS facilites
 * are available.
 *
 * Feel free to contribute...
 */

// Initialize the image reference data with NULL/zero defaults
wxGsImageRefData::wxGsImageRefData()
{
    m_Ok = false;
    m_lockcount = 0;
    m_width = 0;
    m_height = 0;
    m_rowstride = 0;
    m_pixelbytes = 0;
    m_redshift = 0;
    m_greenshift = 0;
    m_blueshift = 0;
    m_imagetype = wxGhostscript::none;
    m_pimage = NULL;
    m_device = NULL;

    // Platform-specific bitmap handles
#ifdef __WXMSW__
    m_hbitmap = NULL;
#elif defined(__WXMAC__)
    m_port = NULL;
#endif
}

// Dispose of the image from memory in a platform-specific way
void wxGsImageRefData::FreeImageData()
{
    if (m_pimage)
    {
#ifdef __WXMSW__
        DeleteObject(m_hbitmap);
#elif defined(__WXOS2__)
#error OS/2 port not yet written. Feel free to contribute.
#elif defined(__WXGTK__)
        if (m_pimage)
            free(m_pimage);
#elif defined(__WXMOTIF__)
#error Motif port not yet written. Feel free to contribute.
#elif defined(__WXX11__)
#error Universal X11 port not yet written. Feel free to contribute.
#elif defined(__WXMAC__)
        DisposeGWorld(m_port);
        if (m_pimage)
            free(m_pimage);
#else
#error Platform not supported.
#endif
        m_pimage = NULL;
        m_Ok = false;
    }
}

IMPLEMENT_DYNAMIC_CLASS(wxGsImage, wxObject)

// Static initialisation of display formats depending on platform facilities
void wxGsImage::InitDisplayFormats()
{
    unsigned int endian, edgefirst, alphadisplay;

    edgefirst = gs::DISPLAY_TOPFIRST;

#if defined(__WXMSW__) || defined(__WXOS2__)
    endian = gs::DISPLAY_LITTLEENDIAN;
#else
    endian = gs::DISPLAY_BIGENDIAN;
#endif

#ifdef __WXMAC__
    alphadisplay = gs::DISPLAY_UNUSED_FIRST;
#else
    alphadisplay = gs::DISPLAY_ALPHA_NONE;
#endif

    ms_colourfmt = gs::DISPLAY_COLORS_RGB | gs::DISPLAY_DEPTH_8 | endian | edgefirst | alphadisplay;
    ms_greyscalefmt = gs::DISPLAY_COLORS_GRAY | gs::DISPLAY_DEPTH_8 | edgefirst;
}

unsigned int wxGsImage::ms_greyscalefmt;
unsigned int wxGsImage::ms_colourfmt;
const wxGsImage wxNullGsImage = wxGsImage();

// The image constructor
wxGsImage::wxGsImage(void *device)
{
    wxGsImageRefData *refData = new wxGsImageRefData;
    m_refData = refData;
    refData->m_device = device;
}

// Initialise the image parameters
void wxGsImage::InitParams(int width, int height, int raster, wxGhostscript::ImageType imagetype)
{
    // Set up image parameters
    M_GSIMGDATA->m_Ok = false;
    M_GSIMGDATA->m_width = width;
    M_GSIMGDATA->m_height = height;
    M_GSIMGDATA->m_rowstride = raster;
    M_GSIMGDATA->m_imagetype = (imagetype == wxGhostscript::colour ? wxGhostscript::colour : wxGhostscript::greyscale);

    // Set up data used for unpacking the image into other formats (i.e. wxImage)
    if (M_GSIMGDATA->m_imagetype == wxGhostscript::colour)
    {
        M_GSIMGDATA->m_pixelbytes = (ms_colourfmt & DISPLAY_ALPHA_MASK ? 4 : 3);
        M_GSIMGDATA->m_redshift = 2;
        M_GSIMGDATA->m_greenshift = 1;
        M_GSIMGDATA->m_blueshift = 0;
    }
    else
    {
        M_GSIMGDATA->m_pixelbytes = 1;
        M_GSIMGDATA->m_redshift = 0;
        M_GSIMGDATA->m_greenshift = 0;
        M_GSIMGDATA->m_blueshift = 0;
    }
}

// Allocate the image in a platform-specic way
void *wxGsImage::AllocateImage(size_t size)
{
    void *mem = NULL;

    wxCHECK_MSG(M_GSIMGDATA, NULL, wxT("not properly initialized"));
    wxCHECK_MSG(M_GSIMGDATA->m_height != 0 && M_GSIMGDATA->m_width != 0, NULL, wxT("invalid bitmap dimensions"));

    M_GSIMGDATA->FreeImageData();
    int height = M_GSIMGDATA->m_height;
#ifdef __WXMSW__
    int width = M_GSIMGDATA->m_width;
#endif
    int rowstride = M_GSIMGDATA->m_rowstride;

//    wxASSERT(size == (size_t)(height * rowstride));

    // This is needed for old versions of GS where too much memory is requested
    if (size > (size_t)(height*rowstride))
    {
        height = (int)size / rowstride;
        if (size > (size_t)(height*rowstride))
            height++;
    }
    if (size < (size_t)(height*rowstride))
        size = (size_t)(height*rowstride);

#ifdef __WXMSW__

    struct
    {
        BITMAPINFOHEADER bmiHeader;
        RGBQUAD          bmiColors[256];
    } bmi;

    for (int i=0; i<256; i++) // palette for greyscale bitmap
    {
        RGBQUAD color;
        color.rgbBlue = i;
        color.rgbGreen = i;
        color.rgbRed = i;
        color.rgbReserved = 0;
        bmi.bmiColors[i] = color;
    }

    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = width;
    bmi.bmiHeader.biHeight = -height;
    bmi.bmiHeader.biSizeImage = size;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = (M_GSIMGDATA->m_imagetype == wxGhostscript::colour ? 24 : 8);
    bmi.bmiHeader.biCompression = BI_RGB;
    bmi.bmiHeader.biClrUsed = 0;
    bmi.bmiHeader.biClrImportant = 0;
    bmi.bmiHeader.biXPelsPerMeter = 0;
    bmi.bmiHeader.biYPelsPerMeter = 0;

    ScreenHDC hdc;
    if (M_GSIMGDATA->m_pimage)
    {
        DeleteObject(M_GSIMGDATA->m_hbitmap);
        M_GSIMGDATA->m_Ok = false;
    }
    M_GSIMGDATA->m_hbitmap = CreateDIBSection(hdc, (BITMAPINFO*)&bmi, DIB_RGB_COLORS, &mem, NULL, 0);

    if(M_GSIMGDATA->m_hbitmap)
        M_GSIMGDATA->m_Ok = true;

#elif defined(__WXOS2__)

// see dpmain.c in GS source for using shared memory, and os2/bitmap.cpp in wx
#error OS/2 port not yet written. Feel free to contribute.

#elif defined(__WXGTK__)

    // Host allocated image - not the most efficient
    mem = malloc(size);
    if (mem)
        M_GSIMGDATA->m_Ok = true;

    gdk_rgb_init();

#elif defined(__WXMOTIF__)

#error Motif port not yet written. Feel free to contribute.

#elif defined(__WXX11__)

// See the porting notes above
#error Universal X11 port not yet written. Feel free to contribute.

#elif defined(__WXMAC__)

    CTabHandle cTable = NULL;
    Rect rect = {0, 0, height, width};

    if (M_GSIMGDATA->m_imagetype != wxGhostscript::colour)
    {
        cTable = GetCTable(40);  // Standard greyscale colour table, sadly inverted
        ColorSpecPtr sa = (*cTable)->ctTable;
        for (int i = 0; i<256; i++)
            InvertColor(&(sa[i].rgb));
    }

    // We would like to be able to have MacOS allocate a native format buffer
    mem = malloc(size);
    if (NewGWorldFromPtr(&M_GSIMGDATA->m_port,
                           (M_GSIMGDATA->m_imagetype == wxGhostscript::colour ? k32ARGBPixelFormat : k8IndexedPixelFormat),
                           &rect, cTable, NULL, 0, (Ptr)mem, (SInt32)rowstride) == noErr)
    {
        PixMapHandle bmappixels = GetGWorldPixMap(M_GSIMGDATA->m_port);
        LockPixels(bmappixels);
        M_GSIMGDATA->m_Ok = true;
    }
    else if (mem)
    {
        free(mem);
        mem = NULL;
    }

    if (M_GSIMGDATA->m_imagetype != wxGhostscript::colour)
        DisposeCTable(cTable);

#else
#error Platform not supported.
#endif

    M_GSIMGDATA->m_pimage = (unsigned char*)mem;
    return mem;
}

// Image locking is needed on Mac as it may move image data around
void wxGsImage::LockImage()
{
    wxCHECK_RET(Ok(), wxT("invalid gs image"));

    if (M_GSIMGDATA->m_lockcount++ == 0)
    {
#ifdef __WXMAC__
        LockPixels(GetGWorldPixMap(M_GSIMGDATA->m_port));
#endif
    }
}

void wxGsImage::UnlockImage()
{
    wxCHECK_RET(Ok(), wxT("invalid gs image"));

    if (M_GSIMGDATA->m_lockcount == 0)
        return;

    if (--(M_GSIMGDATA->m_lockcount) == 0)
    {
#ifdef __WXMAC__
        UnlockPixels(GetGWorldPixMap(M_GSIMGDATA->m_port));
#endif
    }
}

// Get the actual image pixels
unsigned char *wxGsImage::GetImageData()
{
    wxCHECK_MSG(Ok(), NULL, wxT("invalid gs image"));

    LockImage();

#ifdef __WXMAC__
    M_GSIMGDATA->m_pimage = (unsigned char *)GetPixBaseAddr(GetGWorldPixMap(M_GSIMGDATA->m_port));
#endif

    return M_GSIMGDATA->m_pimage;
}

// Copy the image onto a device context, e.g a display window (platform-specific)
void wxGsImage::CopyOnto(const wxDC& dc, wxCoord xdest, wxCoord ydest,
                         wxCoord width, wxCoord height, wxCoord xsrc, wxCoord ysrc)
{
    wxCHECK_RET(dc.Ok(), wxT("invalid device context"));
    wxCHECK_RET(Ok(), wxT("invalid gs image"));

#ifdef __WXMSW__

    HDC cdc = (HDC)dc.GetHDC();
    HDC memdc = CreateCompatibleDC(cdc);

    HGDIOBJ hOldBitmap = SelectObject(memdc, M_GSIMGDATA->m_hbitmap);
    BitBlt(cdc, xdest, ydest, width, height, memdc, xsrc, ysrc, SRCCOPY);
    SelectObject(memdc, hOldBitmap);

    DeleteDC(memdc);

#elif defined(__WXOS2__)
#error OS/2 port not yet written. Feel free to contribute.
#elif defined(__WXGTK__)

    // Limiting to ensure only bitmap is accessed.
    if (xsrc < 0)
    {
        xdest -= xsrc;
        width += xsrc;
        xsrc = 0;
    }
    if (xsrc + width > M_GSIMGDATA->m_width)
        width = M_GSIMGDATA->m_width - xsrc;
    if (ysrc < 0)
    {
        ydest -= ysrc;
        height += ysrc;
        ysrc = 0;
    }
    if (ysrc + height > M_GSIMGDATA->m_height)
        height = M_GSIMGDATA->m_height - ysrc;

    wxWindowDC *wdc = wxStaticCast(&dc, wxWindowDC);

    if (M_GSIMGDATA->m_imagetype == wxGhostscript::colour)
        gdk_draw_rgb_image(wdc->m_window, wdc->m_penGC, xdest, ydest,
                           width, height, GDK_RGB_DITHER_MAX,
                           M_GSIMGDATA->m_pimage + xsrc * 3 + ysrc * M_GSIMGDATA->m_rowstride,
                           M_GSIMGDATA->m_rowstride);
    else
        gdk_draw_gray_image(wdc->m_window, wdc->m_penGC, xdest, ydest,
                            width, height, GDK_RGB_DITHER_MAX,
                            M_GSIMGDATA->m_pimage + xsrc * 3 + ysrc * M_GSIMGDATA->m_rowstride,
                            M_GSIMGDATA->m_rowstride);

#elif defined(__WXMOTIF__)
#error Motif port not yet written. Feel free to contribute.
#elif defined(__WXX11__)
#error Universal X11 port not yet written. Feel free to contribute.
#elif defined(__WXMAC__)

    PixMapHandle bmappixels = GetGWorldPixMap(M_GSIMGDATA->m_port);

    LockImage();

    Rect source = {ysrc, xsrc, height + ysrc, width + xsrc};
    Rect dest   = {ydest, xdest, height + ydest, width + xdest};

    CopyBits(GetPortBitMapForCopyBits(M_GSIMGDATA->m_port),
             GetPortBitMapForCopyBits(GWorldPtr(dc.m_macPort)),
             &source, &dest, srcCopy, NULL);

    UnlockImage();

#else
#error Platform not supported.
#endif
}

// Transfer the image onto a wxBitmap.
wxBitmap wxGsImage::CaptureToBitmap(wxCoord x, wxCoord y, wxCoord width, wxCoord height)
{
    wxBitmap wxbitmap;

    wxCHECK_MSG(Ok(), wxbitmap, wxT("invalid gs image"));

    if (x<0) x=0;
    if (x>=M_GSIMGDATA->m_width) x=M_GSIMGDATA->m_width-1;
    if (width<=0) width = 1;
    if (x+width>M_GSIMGDATA->m_width) width = M_GSIMGDATA->m_width-x;
    if (y<0) y=0;
    if (y>=M_GSIMGDATA->m_height) y=M_GSIMGDATA->m_height-1;
    if (height<=0) height = 1;
    if (y+height>M_GSIMGDATA->m_height) height = M_GSIMGDATA->m_height-y;

    if (wxbitmap.Create(width, height))
    {
        wxMemoryDC bitmapdc;
        bitmapdc.SelectObject(wxbitmap);
        CopyOnto(bitmapdc, 0, 0, width, height, x, y);
    }

    return wxbitmap;
}

// Transfer the image out as a wxImage - useful for saving to a file.
wxImage wxGsImage::CaptureToImage(wxCoord x, wxCoord y, wxCoord width, wxCoord height)
{
    wxImage wximage;

    wxCHECK_MSG(Ok(), wximage, wxT("invalid gs image"));

    if (x<0) x=0;
    if (x>=M_GSIMGDATA->m_width) x=M_GSIMGDATA->m_width-1;
    if (width<=0) width = 1;
    if (x+width>M_GSIMGDATA->m_width) width = M_GSIMGDATA->m_width-x;
    if (y<0) y=0;
    if (y>=M_GSIMGDATA->m_height) y=M_GSIMGDATA->m_height-1;
    if (height<=0) height = 1;
    if (y+height>M_GSIMGDATA->m_height) height = M_GSIMGDATA->m_height-y;

    wximage.Create(width, height);
    if (wximage.Ok())
    {
        unsigned char *dest_image = wximage.GetData();;
        unsigned char *src_image = GetImageData() + M_GSIMGDATA->m_rowstride * y;
        int pixelbytes = M_GSIMGDATA->m_pixelbytes;
        unsigned int redshift = M_GSIMGDATA->m_redshift;
        unsigned int greenshift = M_GSIMGDATA->m_greenshift;
        unsigned int blueshift = M_GSIMGDATA->m_blueshift;

        for (int row = 0; row < height; row++)
        {
            unsigned int *src_pixel = (unsigned int*)(src_image + pixelbytes * x);

            for (wxCoord col = 0; col < width; col++)
            {
                *dest_image++ = (unsigned char)(*src_pixel >> redshift  );
                *dest_image++ = (unsigned char)(*src_pixel >> greenshift);
                *dest_image++ = (unsigned char)(*src_pixel >> blueshift );

                src_pixel += pixelbytes;
            }
            src_image += M_GSIMGDATA->m_rowstride;
        }

        ReleaseImageData();
    }

    return wximage;
}

// Get one image pixel
wxColour wxGsImage::GetPixel(wxCoord x, wxCoord y)
{
    wxCHECK_MSG(Ok(), wxColour(0,0,0), wxT("invalid gs image"));

    unsigned int pixel = (unsigned int)*(GetImageData() +
                         x * M_GSIMGDATA->m_rowstride + y * M_GSIMGDATA->m_pixelbytes);
    ReleaseImageData();

    return wxColour((unsigned char)(pixel >> M_GSIMGDATA->m_redshift  ),
                    (unsigned char)(pixel >> M_GSIMGDATA->m_greenshift),
                    (unsigned char)(pixel >> M_GSIMGDATA->m_blueshift ));
}

// A window class for displaying an image (centralised and with scrollbars when needed)

IMPLEMENT_DYNAMIC_CLASS(wxGsWindow, wxScrolledWindow)

// Attach an image to the window
void wxGsWindow::SetGsImage(const wxGsImage& image)
{
    m_image = image;
    SetVirtualSize(m_image.GetWidth(), m_image.GetHeight());
    SetScrollRate(10, 10);
    Refresh();
}

// Rectangle refresh with adjustments for centralising
void wxGsWindow::RefreshRect(const wxRect& rect)
{
    if (!m_image.Ok())
        return;

    wxRect region(rect);
    int ww, wh, vbX, vbY, xUnit, yUnit;
    int iw = m_image.GetWidth();
    int ih = m_image.GetHeight();
    GetClientSize(&ww, &wh);

    if (iw < ww) region.x += (ww - iw) / 2;
    if (ih < wh) region.y += (wh - ih) / 2;

    // Find Out where the window is scrolled to
    GetViewStart(&vbX, &vbY);
    GetScrollPixelsPerUnit(&xUnit, &yUnit);
    vbX *= xUnit;
    vbY *= yUnit;
    region.x -= vbX;
    region.y -= vbY;

    wxScrolledWindow::RefreshRect(region);
}

// Efficient drawing of only the parts that need updating
void wxGsWindow::OnDraw(wxDC& dc)
{
    if (!m_image.Ok())
        return;

    // Determine the size of the image and the size of the window
    int ww, wh, vbX, vbY, xUnit, yUnit;
    int ibX = 0, ibY = 0;
    int iw = m_image.GetWidth();
    int ih = m_image.GetHeight();
    GetClientSize(&ww, &wh);

    // Find Out where the window is scrolled to
    GetViewStart(&vbX, &vbY);
    GetScrollPixelsPerUnit(&xUnit, &yUnit);
    vbX *= xUnit;
    vbY *= yUnit;

    // Centralise if the image is smaller than the window
    if (iw < ww)
        ibX = (ww - iw) / 2;

    if (ih < wh)
        ibY = (wh - ih) / 2;

    if (ibX!=m_ibx || ibY!=m_iby)
    {
	m_ibx=ibX; m_iby=ibY;
	Refresh();
	return;
     }

    // Get the list of rectangles to be updated
    wxRegionIterator upd(GetUpdateRegion());

    while (upd)
    {
        wxRect dest = upd.GetRect();
#ifndef __WXWIN__
        wxRect desto = dest;
#endif

        // PSt: Adjust the destination if the window is scrolled
	// (it should be done by wxScrolledWindow::DoPrepareDC !!!)
        dest.x += vbX;
        dest.y += vbY;

        // Ensure we don't exceed the bounds of the image when copying
        if (dest.x - ibX + dest.width > iw)
            dest.width = iw - dest.x + ibX;
        if (dest.y - ibY + dest.height > ih)
            dest.height = ih - dest.y + ibY;

#ifndef __WXMSW__
        // GTK
        m_image.CopyOnto(dc, desto, wxPoint(dest.x - ibX, dest.y - ibY));
#else
        // WIN
        m_image.CopyOnto(dc, dest, wxPoint(dest.x - ibX, dest.y - ibY));
#endif

        upd++;
    }
}

// A window class for displaying an image (not centralised)

IMPLEMENT_DYNAMIC_CLASS(wxGsWindowX, wxWindow)

BEGIN_EVENT_TABLE(wxGsWindowX, wxWindow)
    EVT_PAINT(wxGsWindowX::OnPaint)
END_EVENT_TABLE()

// Attach an image to the window
void wxGsWindowX::SetGsImage(const wxGsImage& image)
{
    m_image = image;
    Refresh();
}

// Efficient drawing of only the parts that need updating
void wxGsWindowX::OnPaint(wxPaintEvent& event)
{
    wxPaintDC dc(this);

    if (!m_image.Ok())
        return;

    // Determine the size of the image and the size of the window
    int iw = m_image.GetWidth();
    int ih = m_image.GetHeight();

    // Get the list of rectangles to be updated
    wxRegionIterator upd(GetUpdateRegion());

    while (upd)
    {
        wxRect dest = upd.GetRect();

        // Ensure we don't exceed the bounds of the image when copying
        if (dest.x + dest.width > iw) dest.width = iw - dest.x;
        if (dest.y + dest.height > ih) dest.height = ih - dest.y;

        m_image.CopyOnto(dc, dest, wxPoint(dest.x, dest.y));

        upd++;
    }
}

// Visual tracer - see the GS docs

IMPLEMENT_CLASS(wxGsTraceHost, wxObject)
IMPLEMENT_ABSTRACT_CLASS(wxGsVisualTracer, wxObject)

wxGsVisualTracer::wxGsVisualTracer(wxGsTraceHost& t_host, gs::vd_trace_interface *I1)
{
    host = t_host.m_host;
    scale_x = 0.0; scale_y = 0.0;
    orig_x = 0.0, orig_y = 0.0;
    shift_x = 0.0, shift_y = 0.0;
    get_size_x = static_get_size_x;
    get_size_y = static_get_size_y;
    get_dc = static_get_dc;
    release_dc = static_release_dc;
    erase = static_erase;
    beg_path = static_beg_path;
    end_path = static_end_path;
    moveto = static_moveto;
    lineto = static_lineto;
    curveto = static_curveto;
    closepath = static_closepath;
    circle = static_circle;
    round = static_round;
    fill = static_fill;
    stroke = static_stroke;
    setcolor = static_setcolor;
    setlinewidth = static_setlinewidth;
    text = static_text;
    wait = static_wait;
    m_I1 = I1;
}

double wxGsVisualTracer::static_get_size_x(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    return (vd ? vd->OnGetSizeX() : 0.0);
}

double wxGsVisualTracer::static_get_size_y(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    return (vd ? vd->OnGetSizeY() : 0.0);
}

void wxGsVisualTracer::static_get_dc(gs::vd_trace_interface *I, gs::vd_trace_interface **I1)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
    {
        vd->OnGetDc();
        *I1 = vd->m_I1;
    }
}

void wxGsVisualTracer::static_release_dc(gs::vd_trace_interface *I, gs::vd_trace_interface **I1)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
    {
        vd->OnReleaseDc();
        *I1 = vd->m_I1;
    }
}

void wxGsVisualTracer::static_erase(gs::vd_trace_interface *I, unsigned long c)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnErase(wxColour((unsigned char)(c >> 16), (unsigned char)(c >> 8), (unsigned char)c));
}

void wxGsVisualTracer::static_beg_path(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnBeginPath();
}

void wxGsVisualTracer::static_end_path(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnEndPath();
}

void wxGsVisualTracer::static_moveto(gs::vd_trace_interface *I, double x, double y)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnMoveTo(wxRealPoint(x, y));
}

void wxGsVisualTracer::static_lineto(gs::vd_trace_interface *I, double x, double y)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnLineTo(wxRealPoint(x, y));
}

void wxGsVisualTracer::static_curveto(gs::vd_trace_interface *I, double x0, double y0, double x1, double y1, double x2, double y2)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnCurveTo(wxRealPoint(x0, y0), wxRealPoint(x1, y1), wxRealPoint(x2, y2));
}

void wxGsVisualTracer::static_closepath(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnClosePath();
}

void wxGsVisualTracer::static_circle(gs::vd_trace_interface *I, double x, double y, int r)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnCircle(wxRealPoint(x, y), r);
}

void wxGsVisualTracer::static_round(gs::vd_trace_interface *I, double x, double y, int r)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnRound(wxRealPoint(x, y), r);
}

void wxGsVisualTracer::static_fill(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnFill();
}

void wxGsVisualTracer::static_stroke(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnStroke();
}

void wxGsVisualTracer::static_setcolor(gs::vd_trace_interface *I, unsigned long c)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnSetColor(wxColour((unsigned char)(c >> 16), (unsigned char)(c >> 8), (unsigned char)c));
}

void wxGsVisualTracer::static_setlinewidth(gs::vd_trace_interface *I, unsigned int width)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnSetLineWidth(width);
}

void wxGsVisualTracer::static_text(gs::vd_trace_interface *I, double x, double y, char *ASCIIZ)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnText(wxRealPoint(x, y), wxString(ASCIIZ, wxConvISO8859_1));
}

void wxGsVisualTracer::static_wait(gs::vd_trace_interface *I)
{
    wxGsVisualTracer *vd = wxDynamicCast(I, wxGsVisualTracer);
    if (vd)
        vd->OnWait();
}

void wxGsVisualTracer::LineToMulti(wxRealPoint *pArray, int count)
{
    for (int i = 0; i < count; i++)
        LineTo(Scale(pArray[i]));
}

void wxGsVisualTracer::Bar(wxRealPoint p0, wxRealPoint p1, int width, wxColour c)
{
    SetColor(c);
    SetLineWidth(width);
    BeginPath();
    MoveTo(Scale(p0));
    LineTo(Scale(p1));
    EndPath();
    Stroke();
}

void wxGsVisualTracer::Square(wxRealPoint p, int width, wxColour c)
{
    p = Scale(p);
    SetColor(c);
    SetLineWidth(1);
    BeginPath();
    MoveTo(wxRealPoint(p.x - width, p.y - width));
    LineTo(wxRealPoint(p.x + width, p.y - width));
    LineTo(wxRealPoint(p.x + width, p.y + width));
    LineTo(wxRealPoint(p.x - width, p.y + width));
    LineTo(wxRealPoint(p.x - width, p.y - width));
    EndPath();
    Stroke();
}

void wxGsVisualTracer::Curve(wxRealPoint p0, wxRealPoint p1, wxRealPoint p2, wxRealPoint p3, int width, wxColour c)
{
    SetColor(c);
    SetLineWidth(width);
    BeginPath();
    MoveTo(Scale(p0));
    CurveTo(p1, p2, p3);
    EndPath();
    Stroke();
}

void wxGsVisualTracer::Circle(wxRealPoint p, int radius, wxColour c)
{
    SetColor(c);
    SetLineWidth(1);
    OnCircle(Scale(p), radius);
}

void wxGsVisualTracer::Round(wxRealPoint p, int radius, wxColour c)
{
    SetColor(c);
    SetLineWidth(1);
    OnRound(Scale(p), radius);
}

class wxGsModule: public wxModule
{
DECLARE_DYNAMIC_CLASS(wxGsModule)
public:
    wxGsModule() {}
    bool OnInit() {wxGsImage::InitDisplayFormats(); wxGsDll::InitDefaults(); return TRUE;}
    void OnExit() {}
};

IMPLEMENT_DYNAMIC_CLASS(wxGsModule, wxModule)
