/* Copyright (C) 2003 Nigel A. Hathaway.  All rights reserved.
  
 The contents of this file are subject to the wxWindows License
 Version 3.0 (the "License"); you may not use this file except in
 compliance with the License. You may obtain a copy of the License at
 http://www.wxwindows.org/licence3.txt

 Software distributed under the License is distributed on an
 "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 implied. See the License for the specific language governing
 rights and limitations under the License.

*/

/* $Id: ghostscript.h,v 1.3 2003/09/01 15:49:54 nhathaway Exp $ */

#ifndef ghostscript_H
#define ghostscript_H

#ifdef __GNUG__
    #pragma interface "ghostscript.h"
#endif

#ifdef __WXMAC__
#include <QDOffscreen.h>

// Mac classic only has this function in QuickTime
#if !TARGET_CARBON
EXTERN_API( OSErr )
QTNewGWorldFromPtr(
  GWorldPtr *   gw,
  OSType        pixelFormat,
  const Rect *  boundsRect,
  CTabHandle    cTable,
  GDHandle      aGDevice,
  GWorldFlags   flags,
  void *        baseAddr,
  long          rowBytes)                                     FOURWORDINLINE(0x203C, 0x0020, 0x008B, 0xAAA3);

#define NewGWorldFromPtr QTNewGWorldFromPtr
#endif

#endif

#include "wx/object.h"
#include "wx/dynlib.h"
#include "wx/stream.h"
#include "wx/string.h"
#include "wx/filename.h"
#include "wx/bitmap.h"
#include "wx/image.h"
#include "wx/thread.h"

// Pu GS headers in their own namespace to avoid a clash
namespace gs {
extern "C" {

#include "gs/errors.h"
#include "gs/iapi.h"
#include "gs/vdtrace.h"
#include "gs/gdevdsp.h"

}}  // end of extern C namespace/declarations

class wxGsThread;
class wxGsImage;
class wxGsVisualTracer;

// The Ghostscript DLL class
// -------------------------

#define CALLBACK_POLL_DIVIDER 200

WX_DECLARE_VOIDPTR_HASH_MAP(wxGsImage*, wxGsImageHash);

class wxGsDll : public wxObject
{
DECLARE_CLASS(wxGsDll)
public:
    wxGsDll(wxString GsDllLocation = wxEmptyString, 
            wxString GsLibLocation = wxEmptyString, 
            wxString GsFontsLocation = wxEmptyString);
    ~wxGsDll();

    bool isOK() {return m_OK;}
    double GetVersion() {return m_version;}
    void SetVisualTracer(wxGsVisualTracer& vd);

    friend class wxGhostscript;
    friend class wxGsModule;

protected:
    bool m_OK;
    double m_version;
    wxDynamicLibrary m_dll;
    wxString m_GsLibLocation;
    wxString m_GsFontsLocation;

    // The DLL entry points
    gs::PFN_gsapi_revision gsapi_revision;
    gs::PFN_gsapi_new_instance gsapi_new_instance;
    gs::PFN_gsapi_delete_instance gsapi_delete_instance;
    gs::PFN_gsapi_set_stdio gsapi_set_stdio;
    gs::PFN_gsapi_set_poll gsapi_set_poll;
    gs::PFN_gsapi_set_display_callback gsapi_set_display_callback;
    gs::PFN_gsapi_init_with_args gsapi_init_with_args;
    gs::PFN_gsapi_run_string_begin gsapi_run_string_begin;
    gs::PFN_gsapi_run_string_continue gsapi_run_string_continue;
    gs::PFN_gsapi_run_string_end gsapi_run_string_end;
    gs::PFN_gsapi_run_string_with_length gsapi_run_string_with_length;
    gs::PFN_gsapi_run_string gsapi_run_string;
    gs::PFN_gsapi_run_file gsapi_run_file;
    gs::PFN_gsapi_exit gsapi_exit;
    gs::PFN_gsapi_set_visual_tracer gsapi_set_visual_tracer;

    static wxString ms_defaultDllLocation;
    static void InitDefaults();
};

// Ghostscript instance class
// --------------------------

// Note: currently only one instance can exist at any one time

class wxGhostscript : public wxObject
{
DECLARE_DYNAMIC_CLASS(wxGhostscript)
public:
    enum ImageType {none, colour, greyscale, separations};

    friend class wxGsThread;

    wxGhostscript() {Init();}
    wxGhostscript(wxGsDll *gsdll, const wxArrayString& argarray, ImageType imagetype = none)
                  {Create(gsdll, argarray, imagetype);}
    virtual ~wxGhostscript();

    bool Create(wxGsDll *gsdll, const wxArrayString& argarray, ImageType imagetype = none);

    int GetExitCode() const {return m_exitcode;}

    bool RunString(const wxString& psdata) {return DoOperation(psdata, run);}

    bool RunFile(const wxFileName& fname) {return RunFile(fname.GetFullPath());}
    bool RunFile(const wxString& fname) {return DoOperation(fname, file);}

    bool isOK() {return m_gsdll && m_gsdll->isOK();}
    bool isReady() {return (m_phase == before);}
    bool isSuspended() const {return (m_phase == suspended);}
    bool Suspend();
    bool Resume();
    bool Terminate(int errorcode);

    // Callbacks implemented as virtual functions
    virtual wxString *OnStdin() {return &m_stdinbuf;}
    virtual void OnStdout(wxString text) {}
    virtual void OnStderr(wxString text) {}
    virtual int OnPoll() {return 0;}
    virtual int OnDisplayOpen(wxGsImage& image) {return 0;}
    virtual int OnDisplayPreClose(wxGsImage& image) {return 0;}
    virtual int OnDisplayClose(wxGsImage& image) {return 0;}
    virtual int OnDisplayPreSize(wxGsImage& image) {return 0;}
    virtual int OnDisplaySize(wxGsImage& image) {return 0;}
    virtual int OnDisplaySync(wxGsImage& image) {return 0;}
    virtual int OnDisplayPage(wxGsImage& image, int copies, bool flush) {Suspend(); return 0;}
    virtual int OnDisplayUpdate(wxGsImage& image, wxRect region) {return 0;}

protected:
    enum GsPhase {uninitialized, before, during, suspended, after};
    enum GsOperation {run, file, terminate};
    enum GsCallbackOperation {nocall, onstdin, onstdout, onstderr, onpoll, onopen, 
        onpreclose, onclose, onpresize, onsize, onsync, onpage, onupdate};

    wxGsDll *m_gsdll;
    ImageType m_imagetype;
    void *m_instance;
    int m_exitcode;
    wxGsThread *m_thread;
    bool m_callbackswitch;
    GsPhase m_phase;
    GsPhase m_resume;
    wxGsImageHash m_imagehash;
    wxString m_stdinbuf;
    static int static_pollcnt;

    void Init();

    static gs::display_callback m_display_callback;

    // The real callbacks
    static int GSDLLCALL static_stdin(void *caller_handle, char *buf, int len); 
    static int GSDLLCALL static_stdout(void *caller_handle, const char *str, int len); 
    static int GSDLLCALL static_stderr(void *caller_handle, const char *str, int len);
    static int GSDLLCALL static_poll(void *caller_handle);
    static int static_display_open(void *handle, void *device);
    static int static_display_preclose(void *handle, void *device);
    static int static_display_close(void *handle, void *device);
    static int static_display_presize(void *handle, void *device, int width, int height, int raster, unsigned int format);
    static int static_display_size(void *handle, void *device, int width, int height, int raster, unsigned int format, unsigned char *pimage);
    static int static_display_sync(void *handle, void *device);
    static int static_display_page(void *handle, void *device, int copies, int flush);
    static int static_display_update(void *handle, void *device, int x, int y, int w, int h);
    static void *static_display_memalloc(void *handle, void *device, unsigned long size);
    static int static_display_memfree(void *handle, void *device, void *mem);

    int CheckSuspended(int code);
    bool DoOperation(wxString data, GsOperation op);
    int DoCallback(wxGsImage *image, wxString text, int copies, bool flush, wxRect region, GsCallbackOperation op);
    int DoCallback(wxGsImage *image, wxString text, GsCallbackOperation op)
                  {return DoCallback(image, text, 0, false, wxRect(), op);}
    int DoCallback(wxGsImage *image, int copies, bool flush, GsCallbackOperation op) 
                  {return DoCallback(image, wxEmptyString, copies, flush, wxRect(), op);}
    int DoCallback(wxGsImage *image, wxRect region, GsCallbackOperation op)
                  {return DoCallback(image, wxEmptyString, 0, false, region, op);}
    int DoCallback(wxGsImage *image, GsCallbackOperation op) {return DoCallback(image, wxEmptyString, 0, false, wxRect(), op);}
    bool DoCallbackOp();
    void GsThread();
    void RunOtherThread();
};

// The thread that Ghostscript runs in
// -----------------------------------

// This runs co-operatively and mutually-exclusively with the main thread.
// It has to be like this as WXMAC only supports co-operative threads.
// We run it in a thread, because GS has to be suspended once it has rendered an
//   image, since it wipes the image clean if you let it run to completion.

class wxGsThread : protected wxThread
{
    friend class wxGhostscript;

protected:
    wxGsThread(wxString& inbuf, wxGhostscript *gs, wxThreadKind kind = wxTHREAD_DETACHED) : wxThread(kind), m_gs(gs),
        m_unowned(0), m_op(wxGhostscript::terminate), m_callbackop(wxGhostscript::nocall),
        m_image(NULL), m_copies(0), m_flush(false), m_textin(&inbuf), m_retval(-1), m_errorcode(0) {}
    virtual ExitCode Entry() {m_gs->GsThread(); return 0;}

    wxGhostscript *m_gs;
    wxMutex m_mutex[3];
    int m_unowned;
    wxString m_data;
    volatile wxGhostscript::GsOperation m_op;
    volatile wxGhostscript::GsCallbackOperation m_callbackop;
    wxGsImage *m_image;
    wxString m_text;
    volatile int m_copies;
    volatile bool m_flush;
    wxRect m_region;
    wxString *m_textin;
    volatile int m_retval;
    volatile int m_errorcode;
};

// A class for the image created by Ghostscript. This uses wxWindows' shared data facility
//   so that only one copy of the data exists, even though many instance of the main class
//   exist. The main class contains a reference to the data, and deleting the last instance
//   deletes the data itself.

class wxGsImageRefData : public wxObjectRefData
{
public:
    friend class wxGsImage;
    friend class wxGhostscript;

protected:
    wxGsImageRefData();
    ~wxGsImageRefData() {FreeImageData();};

    void FreeImageData();

    bool m_Ok;
    wxString m_name;
    int m_lockcount;
    int m_width;
    int m_height;
    int m_rowstride;
    int m_pixelbytes;
    unsigned int m_redshift;
    unsigned int m_greenshift;
    unsigned int m_blueshift;
    wxGhostscript::ImageType m_imagetype;
    unsigned char *m_pimage;
    void *m_device;

#ifdef __WXMSW__
    HBITMAP m_hbitmap;
#elif defined(__WXMAC__)
    GWorldPtr m_port;
#endif
};

#define M_GSIMGDATA ((wxGsImageRefData *)m_refData)

class wxGsImage : public wxObject
{
DECLARE_DYNAMIC_CLASS(wxGsImage)
public:

    friend class wxGsDll;
    friend class wxGhostscript;
    friend class wxGsModule;

    wxGsImage() {}
    wxGsImage(const wxGsImage& image) {Ref(image);}

    bool Ok() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_Ok : false);}
    wxString GetName() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_name : wxString(wxEmptyString));}
    void *GetDevice() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_device : NULL);}
    int GetWidth() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_width : 0);}
    int GetHeight() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_height : 0);}
    int GetRowStride() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_rowstride : 0);}
    int GetBytesPerPixel() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_pixelbytes : 0);}
    unsigned int GetRedShift() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_redshift : 0);}
    unsigned int GetGreenShift() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_greenshift : 0);}
    unsigned int GetBlueShift() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_blueshift : 0);}
    wxGhostscript::ImageType GetImageType() const {return (M_GSIMGDATA ? M_GSIMGDATA->m_imagetype : wxGhostscript::none);}
    unsigned char *GetImageData();
    void ReleaseImageData() {UnlockImage();}
    bool IsLocked() const {return (M_GSIMGDATA ? (M_GSIMGDATA->m_lockcount > 0) : true);}

    wxImage CaptureToImage(wxCoord x, wxCoord y, wxCoord width, wxCoord height);
    wxImage CaptureToImage(wxRect r) {return CaptureToImage(r.x, r.y, r.width, r.height);}
    wxBitmap CaptureToBitmap(wxCoord x, wxCoord y, wxCoord width, wxCoord height);
    wxBitmap CaptureToBitmap(wxRect r) {return CaptureToBitmap(r.x, r.y, r.width, r.height);}
    void CopyOnto(const wxDC& dc, wxCoord xdest, wxCoord ydest, 
                  wxCoord width, wxCoord height, wxCoord xsrc, wxCoord ysrc);
    void CopyOnto(const wxDC& dc, wxRect dest, wxPoint src)
        {CopyOnto(dc, dest.x, dest.y, dest.width, dest.height, src.x, src.y);}
    void CopyOnto(const wxDC& dc, wxRect dest)
        {CopyOnto(dc, dest.x, dest.y, dest.width, dest.height, dest.x, dest.y);}
    wxColour GetPixel(wxCoord x, wxCoord y);
    wxColour GetPixel(wxPoint p) {return GetPixel(p.x, p.y);}

    wxGsImage& operator=(const wxGsImage& image)
    {
        if (m_refData != image.m_refData)
            Ref(image);
        return *this;
    }

    bool operator==(const wxGsImage& image) const { return m_refData == image.m_refData; }
    bool operator!=(const wxGsImage& image) const { return m_refData != image.m_refData; }

protected:
    static void InitDisplayFormats();
    static unsigned int ms_colourfmt;
    static unsigned int ms_greyscalefmt;

    wxGsImage(void *device);
    void InitParams(int width, int height, int raster, wxGhostscript::ImageType imagetype);
    void *AllocateImage(size_t size);
    void LockImage();
    void UnlockImage();
    void SetName(const wxString& name) {if(M_GSIMGDATA) M_GSIMGDATA->m_name = name;}
};

extern const wxGsImage wxNullGsImage;

// A scrolled window class to view the image in. This centres the image if the window is 
//   wider than the image, and the area outside the image defaults to the system background.

class wxGsWindow : public wxScrolledWindow
{
DECLARE_DYNAMIC_CLASS(wxGsWindow)
public:
    wxGsWindow() : wxScrolledWindow(), m_image(wxNullGsImage) {m_ibx=0; m_iby=0;}
    wxGsWindow(wxWindow* parent, wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition, 
               const wxSize& size = wxDefaultSize, long style = wxHSCROLL | wxVSCROLL, 
               const wxString& name = wxT("ghostscriptWindow")) : 
        wxScrolledWindow(parent, id, pos, size, style, name), m_image(wxNullGsImage) {m_ibx=0; m_iby=0;}

    void SetGsImage(const wxGsImage& image);
    virtual void RefreshRect(const wxRect& rect);
    virtual void OnDraw(wxDC& dc);

protected:
    wxGsImage m_image;
    int m_ibx;
    int m_iby;
};

class wxGsWindowX : public wxWindow
{
DECLARE_DYNAMIC_CLASS(wxGsWindowX)
public:
    wxGsWindowX() : wxWindow(), m_image(wxNullGsImage) {}
    wxGsWindowX(wxWindow* parent, wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition, 
               const wxSize& size = wxDefaultSize, long style = 0, const wxString& name = wxT("ghostscriptWindow")) : 
        wxWindow(parent, id, pos, size, style, name), m_image(wxNullGsImage) {}

    void SetGsImage(const wxGsImage& image);
    void OnPaint(wxPaintEvent& event);

protected:
    wxGsImage m_image;
private:   
    DECLARE_EVENT_TABLE();
};

// The visual trace system
// -----------------------

// See the GS docs for more details. This provides a simple C++ encapsulation of the
//   existing C API.

// The trace host needs to be sub-classed for the system to work.

class wxGsTraceHost : public wxObject
{
DECLARE_CLASS(wxGsTraceHost)
public:
    wxGsTraceHost(struct gs::vd_trace_host_s *host = NULL) {m_host=host;}

    friend class wxGsVisualTracer;

protected:
    struct gs::vd_trace_host_s *m_host;
};

// An abstract class, with the API callbacks as virtual functions.

class wxGsVisualTracer : protected gs::vd_trace_interface, public wxObject
{
DECLARE_ABSTRACT_CLASS(wxGsVisualTracer)
public:
    friend class wxGsDll;

    wxGsVisualTracer(wxGsTraceHost& host, gs::vd_trace_interface *I1 = NULL);
    virtual ~wxGsVisualTracer() {}

    virtual double OnGetSizeX() = 0;
    virtual double OnGetSizeY() = 0;
    virtual void OnGetDc() = 0;
    virtual void OnReleaseDc() = 0;
    virtual void OnErase(wxColour color) = 0;
    virtual void OnBeginPath() = 0;
    virtual void OnEndPath() = 0;
    virtual void OnMoveTo(wxRealPoint p) = 0;
    virtual void OnLineTo(wxRealPoint p) = 0;
    virtual void OnCurveTo(wxRealPoint p0, wxRealPoint p1, wxRealPoint p2) = 0;
    virtual void OnClosePath() = 0;
    virtual void OnCircle(wxRealPoint p, int radius) = 0;
    virtual void OnRound(wxRealPoint p, int radius) = 0;
    virtual void OnFill() = 0;
    virtual void OnStroke() = 0;
    virtual void OnSetColor(wxColour color) = 0;
    virtual void OnSetLineWidth(unsigned int width) = 0;
    virtual void OnText(wxRealPoint p, wxString s) = 0;
    virtual void OnWait() = 0;

    friend class wxGhostscript;

    void GetDc() {OnGetDc();}
    void ReleaseDc() {OnReleaseDc();}
    double GetSizeUnscaledX() {return OnGetSizeX();}
    double GetSizeUnscaledY() {return OnGetSizeY();}
    double GetSizeScaledX() {return OnGetSizeX()/scale_x;}
    double GetSizeScaledY() {return OnGetSizeY()/scale_y;}
    double GetScaleX() {return scale_x;}
    double GetScaleY() {return scale_y;}
    double GetOriginX() {return orig_x;}
    double GetOriginY() {return orig_y;}
    void SetScale(double scale) {scale_x = scale_y = scale;} 
    void SetScaleXY(double scalex, double scaley) {scale_x = scalex; scale_y = scaley;}
    void SetOrigin(wxRealPoint p) {orig_x = p.x; orig_y = p.y;} 
    void SetShift(wxRealPoint p) {shift_x = p.x; shift_y = p.x;}
    void SetCentralShift() {shift_x = OnGetSizeX()/2; shift_y = OnGetSizeY()/2;} 
    void Erase(wxColour c) {OnErase(c);}
    void BeginPath() {OnBeginPath();}
    void EndPath() {OnEndPath();}
    void MoveTo(wxRealPoint p) {OnMoveTo(Scale(p));}
    void LineTo(wxRealPoint p) {OnLineTo(Scale(p));}
    void LineToMulti(wxRealPoint *pArray, int count);
    void CurveTo(wxRealPoint p0, wxRealPoint p1, wxRealPoint p2) {OnCurveTo(Scale(p0), Scale(p1), Scale(p2));}
    void ClosePath() {OnClosePath();}
    void Bar(wxRealPoint p0, wxRealPoint p1, wxColour c) {Bar(p0, p1, 1, c);}
    void Bar(wxRealPoint p0, wxRealPoint p1, int width , wxColour c);
    void Square(wxRealPoint p, int width, wxColour c);
    void Curve(wxRealPoint p0, wxRealPoint p1, wxRealPoint p2, wxRealPoint p3, wxColour c) {Curve(p0, p1, p2, p3, 1, c);}
    void Curve(wxRealPoint p0, wxRealPoint p1, wxRealPoint p2, wxRealPoint p3, int width, wxColour c);
    void Circle(wxRealPoint p, int radius, wxColour c);
    void Round(wxRealPoint p, int radius, wxColour c);
    void Fill() {OnFill();}
    void Stroke() {OnStroke();}
    void SetColor(wxColour c) {OnSetColor(c);}
    void SetLineWidth(unsigned int w) {OnSetLineWidth(w);}
    void Text(wxRealPoint p, wxString s, wxColour c) {SetColor(c); OnText(p, s);}
    void Wait() {OnWait();}

    wxRealPoint Scale(wxRealPoint p) {p.x = (p.x - orig_x) * scale_x + shift_x; 
                                      p.y = (p.y - orig_y) * scale_y + shift_y; return p;}

private:
    gs::vd_trace_interface *m_I1;

    static double static_get_size_x(gs::vd_trace_interface *I);
    static double static_get_size_y(gs::vd_trace_interface *I);
    static void static_get_dc(gs::vd_trace_interface *I, gs::vd_trace_interface **I1);
    static void static_release_dc(gs::vd_trace_interface *I, gs::vd_trace_interface **I1);
    static void static_erase(gs::vd_trace_interface *I, unsigned long rgbcolor);
    static void static_beg_path(gs::vd_trace_interface *I);
    static void static_end_path(gs::vd_trace_interface *I);
    static void static_moveto(gs::vd_trace_interface *I, double x, double y);
    static void static_lineto(gs::vd_trace_interface *I, double x, double y);
    static void static_curveto(gs::vd_trace_interface *I, double x0, double y0, double x1, double y1, double x2, double y2);
    static void static_closepath(gs::vd_trace_interface *I);
    static void static_circle(gs::vd_trace_interface *I, double x, double y, int r);
    static void static_round(gs::vd_trace_interface *I, double x, double y, int r);
    static void static_fill(gs::vd_trace_interface *I);
    static void static_stroke(gs::vd_trace_interface *I);
    static void static_setcolor(gs::vd_trace_interface *I, unsigned long rgbcolor);
    static void static_setlinewidth(gs::vd_trace_interface *I, unsigned int width);
    static void static_text(gs::vd_trace_interface *I, double x, double y, char *ASCIIZ);
    static void static_wait(gs::vd_trace_interface *I);
};

#endif //ghostscript_H

